# RxIO
# by Eresse <eresse@eresse.net>

# External Includes
require 'openssl'

# RxIO Module
module RxIO

	# I/O Base Module
	module IOBase

		# Chunk Size
		CHUNK_SIZE = 1024

		# Retry Exceptions
		RETRY_EXCEPTIONS = [
			OpenSSL::SSL::SSLErrorWaitReadable,
			IO::EAGAINWaitReadable
		]

		# Process Input:
		# Processes Input for an Endpoint, in the form of a data chunk.
		# @param [Hash] endpoint
		# @param [String] chunk A chunk of data, as received by the socket
		def process_input endpoint, chunk

			# Begin
			begin

				# Pass through Service Handler Module
				@service_handler.filter_input endpoint, chunk

				# Process Messages
				@service_handler.handle_msg endpoint, endpoint[:msgs].shift until endpoint[:msgs].empty?

				# Sub-Process Input
				@service_handler.subprocess_input endpoint if @service_handler.respond_to? :subprocess_input

				# Rescue
			rescue Exception => e

				# Peer Error
				peer_error endpoint, e
			end
		end

		# Read Socket:
		# Attempts to read as many bytes as possible (up to CHUNK_SIZE) from a given socket _s_, passing the data chunks to _process_input_.
		# @param [TCPSocket] s
		def read_sock s

			# Acquire Endpoint for Socket
			e = get_endpoint_for_sock s

			# Check Endpoint
			return unless e

			# OpenSSL doesn't play nice with 'select' because of intermediate buffering.
			# Therefore we loop until we either have an explicit error or a signal to wait.
			done = false
			chunk = ''
			rdstr = ''
			until done || !rdstr
				begin
					rdstr = s.read_nonblock(CHUNK_SIZE)
					chunk << rdstr if rdstr
				rescue Exception => ex
					if RETRY_EXCEPTIONS.include? ex.class; then done = true; else rdstr = nil; end
				end
			end

			# Drop Endpoint & Abort on Error
			return drop_endpoint e unless rdstr

			# Process Input
			process_input e, chunk
		end

		# Write Socket:
		# Attempts to write as many bytes as possible to a given socket _s_ from the associated client's output buffer.
		# @param [TCPSocket] s
		def write_sock s

			# Acquire Endpoint
			e = get_endpoint_for_sock s

			# Check Endpoint
			return unless e

			# Synchronize Endpoint
			e[:lock].synchronize do

				# Write as much as possible
				size = (s.write_nonblock e[:obuf] rescue nil) || 0
				e[:obuf].slice!(0, size) if size > 0
			end
		end

		# Drop Endpoint:
		# Notifies the Service Handler and Parent Implementation (in that order) of a dropped endpoint.
		# @param [Hash] endpoint
		def drop_endpoint endpoint

			# Notify Service Handler
			@service_handler.on_drop endpoint if @service_handler.respond_to? :on_drop

			# Drop Endpoint
			on_drop endpoint
		end

		# Peer Error:
		# Handles an Error from a Peer.
		# @param [Hash] p Endpoint Hash
		# @param [Exception] e Exception
		def peer_error p, _e
			drop_endpoint p
		end
	end
end
